// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
stacksize(1e8);

[m,path]=libraryinfo("toolbox_clarinetlib") ;
DEMOPATH = fullpath(path+"/../demos/") ;
fdrfile = DEMOPATH + "fdr_eagleriver.flt" ;
fdrcoding = [ 32,64,128 ; 16,255,1 ; 8,4,2 ] ;
fdr = CreateFlowDirFromBinary ( fdrfile, "f", fdrcoding ) ;
header = GetFlowDirGridHeader ( fdr ) ;
outlet = XY2rec ( header, [-929387,1901293] ) ;

// Extract streams
t = CreateCellTree ( fdr, outlet ) ;
cellsize = header(3) ;
pixelarea = cellsize^2 ;
supportarea = 1e6 ;
supportweight = supportarea / pixelarea ;
[ STREAMS, StreamOrder ] = MakeStrahlerStreams ( t, supportweight ) ;
DestroyCellTree(t);

// Load elevation raster
elev = CreateRasterFromBinary(DEMOPATH+"elev_eagleriver.flt","f");

// Create and fill distance map
dst = CreateConstantRaster(elev,%nan);
FillDistanceMap ( fdr , dst , outlet , "iterative");
DestroyFlowDirGrid(fdr);

// Get Elevation vs. Distance profile for each stream, and plot it
COLOR = jetcolormap(max(StreamOrder));
for i=1:length(STREAMS)
    D = GetRasterValue(dst,STREAMS(i));
    Z = GetRasterValue(elev,STREAMS(i));
    order = StreamOrder(i);
    plot(D,Z/100,"Color",COLOR(order,:),"LineWidth",order);
end

scf();

for i=1:length(STREAMS)
    D = GetRasterValue(dst,STREAMS(i));
    Z = GetRasterValue(elev,STREAMS(i));
    order = StreamOrder(i);
    plot(D-D($),(Z-Z($))/100,"Color",COLOR(order,:),"LineWidth",order);
end

DestroyFloatRaster(elev);
DestroyFloatRaster(dst);
