// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function [ArealP,ArealT,ArealF] = BuildArealForcings ( PRECIP , ScaleHeight , ResidualP_sample , F0 , MU_Y , VAR_Y , W_sk_precip , S2_sk_precip , W_ok_precip , S2_ok_precip , ..
                                                       TEMPER , LapseRate , ResidualT_sample ,                       W_sk_temper , S2_sk_temper , W_ok_temper , S2_ok_temper , ..
                                                       snow_threshold , XYZ_block )
    
    [ndays,nprecip] = size(ResidualP_sample );
    
    if or( size(XYZ_precip) ~= [nprecip,3] ) then
        error("Les arguments 1 et 2 ont des dimensions incompatibles");
    end

    if or ( size(ResidualP_sample) ~= size(PRECIP) ) then
        error("Les arguments 1 et 2 ont des dimensions incompatibles");
    end    
    
    ntemper = size(TEMPER,'c');
    
    if or ( size(ResidualT_sample) ~= size(TEMPER) ) then
        error("Les arguments 1 et 2 ont des dimensions incompatibles");
    end    
    
    if or( size(XYZ_temper) ~= [ntemper,3] ) then
        error("Les arguments 1 et 2 ont des dimensions incompatibles");
    end

    // Dimensioning output:
        
    ArealT = zeros(ndays,1);
    ArealP = zeros(ndays,1);
    ArealF = zeros(ndays,1);          
    
    //////////////////////////////////////////////////////////////////////////// 
    /// Computing Gauss-Hermite quadrature abcissas and weights once for all ///
    ////////////////////////////////////////////////////////////////////////////
    
    [ugh,wgh] = GaussHermite(10) ;
     
     
    //////////////////////////////////////////////////////////////////////////// 
    ///////// Build precipitation and temperature template at stations /////////
    ////////////////////////////////////////////////////////////////////////////    

    TemplateP_sample = mean(PRECIP,'r') ;

    ResiTemplateP_sample = log(TemplateP_sample) - XYZ_precip(:,3)'/ScaleHeight;
    P0 = exp(mean(ResiTemplateP_sample));
    ResiTemplateP_sample = ResiTemplateP_sample - mean(ResiTemplateP_sample); 
    

    TemplateT_sample = mean(TEMPER,'r') ;

    ResiTemplateT_sample = TemplateT_sample - LapseRate * XYZ_temper(:,3)';
    T0 = mean(ResiTemplateT_sample);
    ResiTemplateT_sample = ResiTemplateT_sample - T0 ;  
    
    //////////////////////////////////////////////////////////////////////////// 
    ///////////////////// Compute template for each block  /////////////////////
    ////////////////////////////////////////////////////////////////////////////
                                  
    Elev = XYZ_block(:,3)';
        
    DriftP_block = P0*exp(Elev/ScaleHeight);

    MU_sk_precip = ResiTemplateP_sample * W_sk_precip ;
 
    TemplateP_block = DriftP_block .* exp( MU_sk_precip + 0.5 * S2_sk_precip );

    
    DriftT_block = T0 + LapseRate * Elev;
        
    MU_sk_temper = ResiTemplateT_sample * W_sk_temper ;
        
    TemplateT_block = DriftT_block + MU_sk_temper ;
             
    //////////////////////////////////////////////////////////////////// 
    /////////////////////////// Loop on days ///////////////////////////
    ////////////////////////////////////////////////////////////////////
    
    
    for j = 1:ndays                                 

        MU_ok_temper = ResidualT_sample(j,:) * W_ok_temper(1:ntemper,:) ;
                                        
        if F0(j)<1 then
                                
            MU_ok_precip = ResidualP_sample(j,:) * W_ok_precip(1:nprecip,:) ;
                                
            if F0(j)>0 then
                   
                Ugh = MU_ok_precip(:) .*. ones(ugh) + sqrt( 2.0 * S2_ok_precip(:)) .*. ugh ;    // matrix [npts,nGH]
                Wgh = ones(MU_ok_precip(:)) .*. wgh ;                             // matrix [npts,nGH]
                  
                // Compute Lambda(Ugh) = ( Fz^(-1) o N ) (U)
                
                Lambda_gh = Anamorph_LN ( Ugh , MU_Y(j)*ones(Ugh) , sqrt(VAR_Y(j))*ones(Ugh) , F0(j)*ones(Ugh) );
                
                // Gauss-Hermite quadrature
                
                Lambda_block = (1/sqrt(%pi))*sum ( Wgh.*Lambda_gh , 'c' );   
                Lambda_block = matrix(Lambda_block,size(MU_ok_precip));                    
           
            else
           
                Lambda_block = exp( MU_Y(j) + sqrt(VAR_Y(j))* MU_ok_precip + VAR_Y(j)*S2_ok_precip/4 );
           
            end
            
        else
            
            Lambda_block = zeros(TemplateP_block);
            
        end

        P_block = TemplateP_block.*Lambda_block;
        ArealP(j) = mean(P_block);
        
        T_block = TemplateT_block + MU_ok_temper ;
        ArealT(j) = mean(T_block);
        
        S_block = sum(P_block(T_block<snow_threshold));
        if(S_block>0)
            ArealF(j) = S_block/ArealP(j);
        end
                    
    end        //    end loop on days
    
endfunction
