// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function LocalT = BuildLocalTemperature2 ( TEMPER , XYZ_temper , LapseRate , VARIO_ResiTemplateT , VARIO_ResidualT , ResidualT_sample , XYZ_estim )
    
    
    if or(isnan(TEMPER)) then
        error("Les series de temperature ne doivent pas comporter de lacunes");
    end
    
    [ndays,ntemper] = size(TEMPER);
    nestim = size(XYZ_estim,'r');
    
    if or( size(XYZ_temper) ~= [ntemper,3] ) then
        error("Les arguments 1 et 2 ont des dimensions incompatibles");
    end

    // Dimensioning output:
    
    LocalT = zeros(ndays,nestim);

    //////////////////////////////////////////////////////////////////////////// 
    ///////////// Compute sample-sample symmetric distance matrices ////////////
    ////////////////////////////////////////////////////////////////////////////
    
    DX = XYZ_temper(:,1)'.*.ones(ntemper,1)-XYZ_temper(:,1).*.ones(1,ntemper);
    DY = XYZ_temper(:,2)'.*.ones(ntemper,1)-XYZ_temper(:,2).*.ones(1,ntemper);
    D_tt = sqrt(DX.^2+DY.^2);           
        
    //////////////////////////////////////////////////////////////////////////// 
    /////////// Structural functions and associated kriging matrices ///////////
    ////////////////////////////////////////////////////////////////////////////            
      
    K_tt_template = CovarianceMatrix ( D_tt , VARIO_ResiTemplateT ) ;
    cholK_tt_template = KrigCholesky ( K_tt_template ) ;    
           
    Gamm_tt = CovarianceMatrix ( D_tt , VARIO_ResidualT ) ;
    Gamm_tt = 1-Gamm_tt;
    FactVar = VARIO_ResidualT(2);
    MatKrig_tt = [-Gamm_tt,ones(ntemper,1);ones(1,ntemper),0];
    [LU_tt,IPIV_tt] = KrigBunchKaufman ( MatKrig_tt ) ;
            
    //////////////////////////////////////////////////////////////////////////// 
    ///////// Build precipitation and temperature template at stations /////////
    ////////////////////////////////////////////////////////////////////////////            
       
    TemplateT_sample = mean(TEMPER,'r') ;

    ResiTemplateT_sample = TemplateT_sample - LapseRate * XYZ_temper(:,3)';
    T0 = mean(ResiTemplateT_sample);
    ResiTemplateT_sample = ResiTemplateT_sample - T0 ;  
     
    //////////////////////////////////////////////////////////////////////////// 
    ////////////////////////// Compute local forcings //////////////////////////
    ////////////////////////////////////////////////////////////////////////////
      
    X_estim = XYZ_estim(:,1);
    Y_estim = XYZ_estim(:,2);
    
    DX = XYZ_temper(:,1).*.ones(X_estim') - ones(ntemper,1).*.X_estim' ;
    DY = XYZ_temper(:,2).*.ones(Y_estim') - ones(ntemper,1).*.Y_estim' ;
    D_te = sqrt(DX.^2+DY.^2);                  
                            
    Elev = XYZ_estim(:,3);
        
    DriftT_estim = T0 + LapseRate * Elev;
    
    K_te_template = CovarianceMatrix ( D_te , VARIO_ResiTemplateT ) ;
    
    Gamm_te = CovarianceMatrix ( D_te , VARIO_ResidualT ) ;
    Gamm_te = 1-Gamm_te;
    
    [ MU_sk , S2_sk ] = SimpleKriging_fast ( cholK_tt_template , ResiTemplateT_sample' , K_te_template , VARIO_ResiTemplateT(2) ) ;
    TemplateT_estim = DriftT_estim + MU_sk ;
                
    //////////////////////////////////////////////////////////////////// 
    /////////////////////////// Loop on days ///////////////////////////
    ////////////////////////////////////////////////////////////////////
    
    [ W_ok , S2_ok ] = OrdinaryKriging_fast ( LU_tt , IPIV_tt , Gamm_te , FactVar ) ;
    
    for j = 1:ndays                                 
                                
            MU_ok = ResidualT_sample(j,:) * W_ok(1:ntemper,:) ;
                                
            T_estim = TemplateT_estim + MU_ok ;
            LocalT(j,:) = matrix(T_estim,1,nestim);
                    
    end        //    end loop on days
    
endfunction
