// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function [ VARIO_ResidualT , ResidualT_sample ] = DailyT_variography ( TEMPER , XYZ_temper )
        
    
    if or(isnan(TEMPER)) then
        error("Les series de temperature ne doivent pas comporter de lacunes");
    end
    
    [ndays,ntemper] = size(TEMPER);
    
    if or( size(XYZ_temper) ~= [ntemper,3] ) then
        error("Les arguments 1 et 2 ont des dimensions incompatibles");
    end

    //////////////////////////////////////////////////////////////////////////// 
    ///////////// Compute sample-sample symmetric distance matrices ////////////
    ////////////////////////////////////////////////////////////////////////////
    
    DX = XYZ_temper(:,1)'.*.ones(ntemper,1)-XYZ_temper(:,1).*.ones(1,ntemper);
    DY = XYZ_temper(:,2)'.*.ones(ntemper,1)-XYZ_temper(:,2).*.ones(1,ntemper);
    D_tt = sqrt(DX.^2+DY.^2);
                        
    //////////////////////////////////////////////////////////////////////////// 
    ///////// Build precipitation and temperature template at stations /////////
    ////////////////////////////////////////////////////////////////////////////            
       
    TemplateT_sample = mean(TEMPER,'r') ;

    //////////////////////////////////////////////////////////////////////////// 
    //         Compute residuals to templates at gauges for each day :        //
    // additive for temperature and multiplicative, anamorphosed for rainfall // 
    ////////////////////////////////////////////////////////////////////////////

    ResidualT_sample = TEMPER - TemplateT_sample.*.ones(ndays,1) ;
    
    //////////////////////////////////////////////////////////////////////////// 
    //////// Variographie des résidus à l'ébauche (intermède musical 2) ////////
    ////////////////////////////////////////////////////////////////////////////

    h = zeros(ntemper*(ntemper-1)/2,1);
    g = zeros(h);    
    i = 0;
    E2 = 0;
                
    for t1=1:(ntemper-1)       
        for t2=(t1+1):ntemper
            i=i+1;
            h(i) = D_tt(t1,t2);
            accr = ResidualT_sample(:,t1) - ResidualT_sample(:,t2) ;
            g(i) = 0.5*mean(accr.^2) ;
            E2 = E2 + sum(accr.^2) ;
        end       
    end
    
    Sill = E2/(ntemper*(ntemper-1)*ndays) ; // variance(ResidualT_sample - mean(ResidualT_sample,'c').*.ones(1,ntemper)) ;                       
    Nugget = 0 ;
    X0 = [ 0 ] ;
    [fopt,Xopt,gropt] = leastsq ( 0 , list ( VarioDiff_wo_nugget , Nugget , Sill , 'spherical' , h , g ) , X0 , 'ar' , 100 , 50 );        
    Range = max(h) * 0.5 * ( 1 + tanh ( Xopt(1) ) );
    
    VARIO_ResidualT = list ( "spherical" , Sill , Range , Nugget );
        
endfunction
