// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function [DriftP_block,TemplateP_block,DailyScalingFactor,LocalP_block] = DecompLocalPrecip ( XYZ_precip , ScaleHeight , TemplateP_sample , ResidualP_sample , F0 , MU_Y , VAR_Y , W_sk_precip , S2_sk_precip , W_ok_precip , S2_ok_precip , XYZ_block )
    
    [ndays,nprecip] = size(ResidualP_sample );   

    // Dimensioning output:

    DailyScalingFactor = zeros(ndays,size(XYZ_block,'r'));
        
    //////////////////////////////////////////////////////////////////////////// 
    /// Computing Gauss-Hermite quadrature abcissas and weights once for all ///
    ////////////////////////////////////////////////////////////////////////////
    
    [ugh,wgh] = GaussHermite(10) ;
     
    //////////////////////////////////////////////////////////////////////////// 
    ///////// Build precipitation and temperature template at stations /////////
    ////////////////////////////////////////////////////////////////////////////    

    ResiTemplateP_sample = log(TemplateP_sample) - XYZ_precip(:,3)'/ScaleHeight;
    P0 = exp(mean(ResiTemplateP_sample));
    ResiTemplateP_sample = ResiTemplateP_sample - mean(ResiTemplateP_sample); 
       
    //////////////////////////////////////////////////////////////////////////// 
    ///////////////////// Compute template for each block  /////////////////////
    ////////////////////////////////////////////////////////////////////////////
                                  
    Elev = XYZ_block(:,3)';
        
    DriftP_block = P0*exp(Elev/ScaleHeight);

    MU_sk_precip = ResiTemplateP_sample * W_sk_precip ;
 
    TemplateP_block = DriftP_block .* exp( MU_sk_precip + 0.5 * S2_sk_precip );    
                
    //////////////////////////////////////////////////////////////////// 
    /////////////////////////// Loop on days ///////////////////////////
    ////////////////////////////////////////////////////////////////////
    
    
    for j = 1:ndays                                 
                                        
        if F0(j)<1 then
                                
            MU_ok_precip = ResidualP_sample(j,:) * W_ok_precip(1:nprecip,:) ;
                                
            if F0(j)>0 then
                   
                Ugh = MU_ok_precip(:) .*. ones(ugh) + sqrt(S2_ok_precip(:)) .*. ugh ;    // matrix [npts,nGH]
                Wgh = ones(MU_ok_precip(:)) .*. wgh ;                             // matrix [npts,nGH]
                  
                // Compute Lambda(Ugh) = ( Fz^(-1) o N ) (U)
                
                Lambda_gh = Anamorph_LN ( Ugh , MU_Y(j)*ones(Ugh) , sqrt(VAR_Y(j))*ones(Ugh) , F0(j)*ones(Ugh) );
                
                // Gauss-Hermite quadrature
                
                Lambda_block = (1/sqrt(%pi))*sum ( Wgh.*Lambda_gh , 'c' );   
                Lambda_block = matrix(Lambda_block,size(MU_ok_precip));                    
           
            else
           
                Lambda_block = exp( MU_Y(j) + sqrt(VAR_Y(j))* MU_ok_precip + VAR_Y(j)*S2_ok_precip/4 );
           
            end
            
        else
            
            Lambda_block = zeros(TemplateP_block);
            
        end

        DailyScalingFactor(j,:)=Lambda_block;
        LocalP_block(j,:) = TemplateP_block.*Lambda_block;
                            
    end        //    end loop on days
    
endfunction
