// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function ixPivots = DouglasPeucker ( Polyline , tol , offset )
    
    // Identification du point pivot

    n = size(Polyline,'r');    
    Polyline = Polyline - mean(Polyline,'r').*.ones(n,1);
    u0 = Polyline($,:)-Polyline(1,:) ;
    
    if u0==0 then
        ixPivots = [1 n] + offset ;
        return
    end
    
    L0 = norm(u0);
    u0 = u0/L0;
    u = Polyline - Polyline(1,:).*.ones(n,1) ;
    ps = sum(u.*[u0.*.ones(n,1)],'c');
    ps(ps<0) = 0;
    ps(ps>=L0) = L0;
    projec = u0.*.ps;
    ortho = u - projec ;
    ortho($) = 0;
    
    d = sqrt(sum(ortho.^2,'c'));
    [dmax,pivot] = max(d);
    
    if dmax > tol
        
        // Ajout du pivot et appel récursif sur les 2 sous-polylignes     
    
        offset1 = offset ;
        ixP1 = DouglasPeucker ( Polyline(1:pivot,:) , tol , offset1 );
        offset2 = offset+pivot-1 ;
        ixP2 = DouglasPeucker ( Polyline(pivot:$,:) , tol , offset2 );
        
        ixPivots = unique([ ixP1 , ixP2 ]) ;
    
    else
    
        ixPivots = [1 n] + offset ;
        
    end  
    
endfunction
