// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function Z = EstimZ_GH ( Ysk , Ssk , f0 , mlog , slog , n) ;
    
    // estime E [ Z(x0) = h(Y(x0)) | Y(x1) ... Y(xn) ]
    // où Y est l'anamorphose gaussienne de Z par la fonction h = Fz^(-1) o N
    // Fz    : densité de la loi log-normale de paramètres mlog, slog avec une valeur atomique f0 en 0
    // N     : densité de la loi normale N(0,1)
    // n     : entier, ordre pour la quadrature de Gauss-Hermite
   
    if(f0>=1)
        Z = zeros(Ysk);
        return;
    end
    
     
    [xgh,wgh] = GaussHermite(n) ;    // Calcul des poids pour la quadrature
    
    npts = length(Ysk);
    Ugh = Ysk(:) .*. ones(xgh) + sqrt(2)*Ssk(:) .*. xgh ;    // matrice [npts,n]
    Wgh = ones(Ysk(:)) .*. wgh ;                     // matrice [npts,n]
      
    // Calcul de h(Ugh) = ( Fz^(-1) o N ) (U)
    
    Zgh = Anamorph_LN ( Ugh , mlog , slog , f0 );
    
    // Calcul de la quadrature
    
    Z = (1/sqrt(%pi))*sum ( Wgh.*Zgh , 'c' );   
    Z = matrix(Z,size(Ysk));
    
endfunction
