// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function Noeuds = OptimisationMaillage ( fdr , wraster , exu , target , varargin )
    
    if length(varargin,'*')>0 then
        id_cnstr = varargin(1);
        id_cnstr = unique([id_cnstr(:);exu]);
    else
        id_cnstr = exu;
    end
    
    // Construction de la carte d'accumulation du raster de poids
    
    accw = CreateConstantRaster(wraster,%nan);
    Accumulate ( fdr , wraster , exu , accw );
    
    // Construction de la carte de distance à l'exutoire 
    
    pk = CreateConstantRaster(wraster,%nan);
    FillDistanceMap ( fdr,pk,exu);
    
    // Estimation d'un premier ensemble de noeuds
    
    NODES = SegmentationTopo(fdr,wraster,exu,target,id_cnstr);
    NODES = NODES(:);
    
    is_cnstr = zeros(NODES)>0 ;
    
    // Elimination des points de contrainte non-effectifs 
    
    for i=1:length(NODES)
        if ~isempty( find(id_cnstr==NODES(i)) ) then
            is_cnstr(i) = %T;
        end
    end
    
    noptim = sum(~is_cnstr);
    OPTIMNODES = NODES(~is_cnstr);
    
    // Extraction des filets le long desquels on optimisera la position des noeuds
    
    FLOWLINES = list();
    pk0 = GetRasterValue(pk,OPTIMNODES);   
    bornes = zeros(length(pk0),2);
    
    for i=1:noptim
        FlowLine_i = MakeFlowLineThroughPoint(fdr,accw,exu,OPTIMNODES(i));
        pk_i = GetRasterValue(pk,FlowLine_i);
        FLOWLINES(i) = flipdim([pk_i(:),FlowLine_i(:)],1);
        bornes(i,2) = pk_i(1);
            
        // Réduction des bornes [pkmin,pkmax] entre lesquels déplacer le noeud sur le filet,
        // en fonction de la présence d'autres noeuds sur ce même filet
        
        for j=1:length(NODES)
            
            ix = find(FlowLine_i==NODES(j));
            
            if NODES(j)~=OPTIMNODES(i) & ~isempty(ix) then
                if pk_i(ix)<pk0(i) then
                    bornes(i,1) = max ( bornes(i,1) , pk_i(ix) );
                end
                if pk_i(ix)>pk0(i) then
                    bornes(i,2) = min ( bornes(i,2) , pk_i(ix) );
                end
            end
            
        end
        
    end
    
    pause
    
    // Optimisation
    
    [f,xopt, gropt] = leastsq(list(EvalMaillage,FLOWLINES, fdr ,wraster,NODES(is_cnstr)),"b",bornes(:,1),bornes(:,2),pk0,"ar",50,8,1e-3,1e-3,30*5,imp=2);    
        
endfunction
