// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function [ Z_estim , V_estim ] = OrdinaryKriging ( XY_sample , Z_sample , XY_estim , VARIO )
    
    // VARIO : list(type,Sill,Range,Nugget)
    
    n_sample = length(Z_sample) ;      
    Z_sample = Z_sample(:)' ;
    
    //////////////////////////////////////////////////////////////////////////// 
    /////////////////// Variogramme des points échantillonnés //////////////////
    ////////////////////////////////////////////////////////////////////////////     
    
    X_sample = XY_sample(:,1);
    Y_sample = XY_sample(:,2);
          
    DX = X_sample'.*.ones(X_sample)-X_sample.*.ones(X_sample');
    DY = Y_sample'.*.ones(Y_sample)-Y_sample.*.ones(Y_sample');
    D = sqrt(DX.^2+DY.^2);
          
    type_vario = convstr ( part(VARIO(1),1:1) , 'l' );
    Sill = VARIO(2);
    Range = VARIO(3);
    Nugget = VARIO(4);
          
    select type_vario
    case 'e' then    // exponential
        G = Nugget + ( Sill - Nugget ) * (1-exp(-D/Range));
    case 's' then    // spherical
        G = Sill * ones(D);
        G(D<Range) = Nugget + ( Sill - Nugget ) * [ 1.5*D(D<Range)/Range-0.5*(D(D<Range)/Range).^3 ] ;
    case 'e' then    // gaussian
        G = Nugget + ( Sill - Nugget ) * (1-exp(-[D/Range].^2));
    end
    
    G = G / Sill // variogramme pré-conditionné
    
    //////////////////////////////////////////////////////////////////////////// 
    ////////////////////// Estimation aux points XY_estim //////////////////////
    ////////////////////////////////////////////////////////////////////////////    

    X_estim = XY_estim(:,1);
    Y_estim = XY_estim(:,2);
    n_estim = length(X_estim);

    // Calcul de covariance entre les points echantillonnes et la localité XY_estim(i,:)
    // Matrice [n_sample,n_estim]
            
    DX = X_sample.*.ones(X_estim') - ones(X_sample).*.X_estim' ;
    DY = Y_sample.*.ones(Y_estim') - ones(Y_sample).*.Y_estim' ;
    D = sqrt(DX.^2+DY.^2);

    select type_vario
    case 'e' then    // exponential
        Gi = Nugget + ( Sill - Nugget ) * (1-exp(-D/Range));
    case 's' then    // spherical
        Gi = Sill * ones(D);
        Gi(D<Range) = Nugget + ( Sill - Nugget ) * [ 1.5*D(D<Range)/Range-0.5*(D(D<Range)/Range).^3 ] ;
    case 'e' then    // gaussian
        Gi = Nugget + ( Sill - Nugget ) * (1-exp(-[D/Range].^2));
    end
    
    Gi = Gi / Sill ;
          
    // Systeme de krigeage ordinaire
          
    A = [ G , -ones(n_sample,1) ; ones(1,n_sample) , 0 ] ;
    B = [ Gi ; ones(1,n_estim) ] ;
            
    lambda = A\B ;
           
    Z_estim = Z_sample * lambda(1:n_sample,:) ;
    Z_estim = Z_estim' ;
    
    V_estim = zeros(Z_estim);
    
    for i=1:length(X_estim)    
        V_estim(i) = Gi(:,i)' * lambda(1:n_sample,i) - lambda($,i)  ;                 
    end         
    
    V_estim = Sill * V_estim ;
           
endfunction
