// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function [ logai , logpi , dyi ] = PrepareSet ( Ai , ni , alpha , npoints, IC )
          
        if ni<length(Ai) then
               error("n doit être au moins égal au nombre d''éléments de A");
        end
  
        Ai = gsort(Ai,'g','i');
        Oi = [1:length(Ai)]' + ni-length(Ai) ;
        Fi = [Oi-0.31]/(ni+1-2*0.31);    // Fréquence empirique

        // On réduit la taille du jeu en prenant 1 point par plage logarithmique de 1-F
              
        log1_Fg = [0:npoints] * [ log(1-Fi(1)) - log(1-Fi($)) ]/npoints + log(1-Fi($));
        ix = (ni+1-2*0.31) * (1-exp(log1_Fg)) + 0.31 ;
        ix = unique(round(ix))-ni+length(Ai);
                
        Ai = Ai(ix);
        Oi = Oi(ix);
        Fi = Fi(ix);
                       
        // Calcul des intervalles de confiance au niveau (1-alpha) en chaque point retenu
        
        Li = zeros(Ai);
        Hi = zeros(Ai);
        
        if IC<=0 then
            
                // Méthode analytique par récursion (lent)
                
                for j=1:length(Ai)
                    
                    tic;
                    mprintf("        Calcul de l''intervalle de confiance %d:%d (F=%.3e)... ",Oi(j),ni,Fi(j))
                    [L,H,cl] = PointWiseConfidence(ni,Oi(j),Fi(j),alpha,-log(1-alpha),100,100);
                    Li(j) = L;
                    Hi(j) = H;
                    t=toc();
                    mprintf("%.1f s\n",t)
                        
                end

        else                  
            
                // Méthode par simulation directe (plus rapide !!)
                    
                ns = IC;    // nombre d'échantillons de taille ni
                OrderStatSample = zeros(length(Ai),ns);
                
                for j=1:ns
                    Sample = rand(ni,1);
                    Sample = gsort(Sample,'g','i');
                    OrderStatSample(:,j) = Sample(Oi);
                end
            
                OrderStatSample = gsort(OrderStatSample,'c','i');
                [Ix,W] = InterpQuantile ([alpha/2 , 1-alpha/2],ns);
                Li = OrderStatSample(:,Ix(1,1))*W(1,1)+OrderStatSample(:,Ix(1,2))*W(1,2);
                Hi = OrderStatSample(:,Ix(2,1))*W(2,1)+OrderStatSample(:,Ix(2,2))*W(2,2);
        end        
        
        // Jeu transformé et habillé d'une incertitude :

//        xi = log(Ai) + gam*log(Di);
          logai = log(Ai);
//        yi = bet*log(Ai) + log(1-Fi);
          logpi = log(1-Fi);
          dyi = 0.5*( log(1-Li) - log(1-Hi) );
       
    
endfunction
