// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function [Wtrial,biais,cv_opt] = ProposeMesh (fdr,exu,optim_type,cible,varargin)

//------------------------------------------------------------------------------
// Lecture du type de cible choisi : poids moyen ou nombre de cellules
         
    select convstr(optim_type,"l")
        case "number" then
            with_fixed_number = %T; 
            with_fixed_weight = %F;
        case "weight" then
            with_fixed_number = %F; 
            with_fixed_weight = %T;            
        else
            errormsg = sprintf("Argument 3 incorrect (%s) : doit être ''number'' ou ''weight''\n",optim_type);
            error(errormsg);
    end
     
//------------------------------------------------------------------------------
// Lecture des arguments optionnels

    nargs = length(varargin,'*');
    CNT = [];
    useweight = %F ;
        
    if nargs>0 then
        
        if modulo(nargs,2)~=0
            error("Syntaxe incorrecte");
        end

        // Lecture des clés et des arguments associés
        
        for i=2*[1:(nargs/2)]
            
            if convstr(varargin(i-1),"l")=="constrained" then
                CNT = varargin(i);
            end

            if convstr(varargin(i-1),"l")=="weight_raster" then
                useweight = %T;
                wloc = varargin(i);
            end
            
        end
        
    end

//------------------------------------------------------------------------------
// Création des rasters temporaires nécessaires au calcul
   
    mprintf("Calcul de la carte d''accumlation des poids...")

    Header = GetFlowDirGridHeader (fdr);
    wcum = CreateFloatRaster("cumulated weight",Header(1),Header(2),Header(3),Header(4),Header(5),Header(6),Header(7));    
    SetRasterToNODATA(wcum);
               
    if ~useweight then
        wloc = CreateFloatRaster("local weight",Header(1),Header(2),Header(3),Header(4),Header(5),Header(6),Header(7)); 
        SetRasterToConstant(wloc,1);
    end
   
    FillAccumulationMap(fdr,exu,wcum,wloc);
    
    mprintf(" fait.\n")

    mprintf("Calcul de la carte de distance...")
    dist = CreateConstantRaster(wcum,%nan);
    FillDistanceMap(fdr,dist,exu);
    mprintf(" fait.\n")
    
//------------------------------------------------------------------------------
// Création du semis de noeuds initial : minimisation de l'écart à la cible
    
    Wtot = GetRasterValue ( wcum , exu );
    
    if with_fixed_number then
        Wcible = Wtot / cible ;
    else
        Wcible = cible ;
    end
       
    Winf = 0 ;
    Wsup = 2*Wcible;
    tol = 1e-3;
    errmin = %inf;
    
    mprintf("\n Minimisation du biais sur le poids...\n")
    
    while (Wsup-Winf)/Wcible > tol 
        
        W = 0.5*(Winf+Wsup);
        
        [NODES,WEIGHTS] = SegmentationTopo(fdr,wloc,exu,W,CNT);
        Weff = mean(WEIGHTS);
        err = (Weff -  Wcible)/Wcible ; 
                
        if err<0 then
            Winf = W ;
        else
            Wsup = W ;
        end
        
        if ( abs(err) < errmin ) then
            mprintf(">> biais %.1f %%\n",100*err)
            Wopt = W;
            errmin = abs(err);
        end
        
    end
    
    n=6;
    r = 1.5;
    l1 = 0.5*(1-r)/(1-r^n);
    R=r.^[0:(n-1)];
    fact = cumsum(l1*R);
    fact = 2*[-0.5,-fact((n-1):-1:1),0,fact];
    fact = (3/2).^fact;

    Wtrial = Wopt * fact ;   

//------------------------------------------------------------------------------
// Déplacement des noeuds du semis initial pour minimiser la variance des poids

    cv_opt = zeros(Wtrial);
    biais = zeros(Wtrial);

    for t=1:length(Wtrial)

            [NODES,WEIGHTS] = SegmentationTopo(fdr,wloc,exu,Wtrial(t),CNT);
            isCNT = zeros(NODES);
            for i=1:length(CNT)
                isCNT(NODES==CNT(i)) = 1;
            end   
            isCNT(NODES==exu) = 1;
                
            cv0 = stdev(WEIGHTS/mean(WEIGHTS));
            biais(t) = (mean(WEIGHTS)-Wcible)/Wcible;
            cv_opt(t) = cv0;
            cv_min = 0;
            nitermax = 10;
            niter = 0;
            DAMP = 0.4:0.1:0.6;//0:0.1:0.9;
            
            msg = sprintf("CONFIG %d : Biais = %.1f %%",t,100*biais(t));
            mprintf("\n%s\n",msg)
            mprintf("Minimisation du coefficient de variation des poids...\n")
            
            while (cv_min < cv_opt(t) & niter < nitermax)
        
                cv = zeros(DAMP);
                niter = niter+1;
           
                for i=1:length(DAMP)
                    [NEWNODES,NEWWEIGHTS] = sci_OptimMeshVariance (fdr,wcum,dist,NODES,isCNT,DAMP(i));
                    cv(i)=stdev(NEWWEIGHTS/mean(NEWWEIGHTS));
                end
                
                [cv_min,iopt]=min(cv);
                       
                if cv_min<cv_opt then
                    cv_opt(t) = cv_min;
                    [NODES,WEIGHTS] = sci_OptimMeshVariance (fdr,wcum,dist,NODES,isCNT,DAMP(iopt));
                    mprintf(">> coeff.var %.1f %%\n",100*cv_opt(t))
                end
            
            end

    end

//------------------------------------------------------------------------------
// Destruction des rasters temporaires

    if ~useweight then
        DestroyFloatRaster(wloc);
    end

    DestroyFloatRaster(dist);
    DestroyFloatRaster(wcum);

endfunction
