// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function RasterOK ( handle , XY_sample , Z_sample , VARIO , varargin )
    
    // VARIO : list(type,Sill,Range,Nugget)
    
    if ~isempty(varargin) then
        handlevar = varargin(1);
        estimvar = %T;
    else
        estimvar = %F;
    end
    
    n_sample = length(Z_sample) ;      
    Z_sample = Z_sample(:)' ;
    
    //////////////////////////////////////////////////////////////////////////// 
    /////////////////// Variogramme des points échantillonnés //////////////////
    ////////////////////////////////////////////////////////////////////////////     
    
    X_sample = XY_sample(:,1);
    Y_sample = XY_sample(:,2);
          
    DX = X_sample'.*.ones(X_sample)-X_sample.*.ones(X_sample');
    DY = Y_sample'.*.ones(Y_sample)-Y_sample.*.ones(Y_sample');
    D = sqrt(DX.^2+DY.^2);
          
    type_vario = convstr ( part(VARIO(1),1:1) , 'l' );
    Sill = VARIO(2);
    Range = VARIO(3);
    Nugget = VARIO(4);
          
    select type_vario
    case 'e' then    // exponential
        G = Nugget + ( Sill - Nugget ) * (1-exp(-D/Range));
    case 's' then    // spherical
        G = Sill * ones(D);
        G(D<Range) = Nugget + ( Sill - Nugget ) * [ 1.5*D(D<Range)/Range-0.5*(D(D<Range)/Range).^3 ] ;
    case 'e' then    // gaussian
        G = Nugget + ( Sill - Nugget ) * (1-exp(-[D/Range].^2));
    end
    
    G = G / Sill // variogramme pré-conditionné
    
    //////////////////////////////////////////////////////////////////////////// 
    //////////////////// Estimation aux points de la grille ////////////////////
    ////////////////////////////////////////////////////////////////////////////    

    Header = GetFloatRasterHeader(handle);
    ncols = Header(1);
    nrows = Header(2);

    Z_estim = zeros(1,ncols);
    Z_variance = zeros(1,ncols);

    for r=0:(nrows-1)
        
        XY_estim = rec2XY ( Header, r*ncols + [1:ncols] - 1 ) ;
                
        for i=1:ncols
             
            // Calcul du variogramme entre les points echantillonnes et la localité XY_estim(i,:)
            
            D = sqrt ( ( X_sample - XY_estim(i,1) ).^2 + ( Y_sample - XY_estim(i,2) ).^2 );
    
            select type_vario
            case 'e' then    // exponential
                Gi = Nugget + ( Sill - Nugget ) * (1-exp(-D/Range));
            case 's' then    // spherical
                Gi = Sill * ones(D);
                Gi(D<Range) = Nugget + ( Sill - Nugget ) * [ 1.5*D(D<Range)/Range-0.5*(D(D<Range)/Range).^3 ] ;
            case 'e' then    // gaussian
                Gi = Nugget + ( Sill - Nugget ) * (1-exp(-[D/Range].^2));
            end
          
            Gi = Gi / Sill ;
          
            // Systeme de krigeage ordinaire
          
            A = [ G , -ones(n_sample,1) ; ones(1,n_sample) , 0 ] ;
            B = [ Gi ; 1 ] ;
            
            lambda = A\B ;

            Z_estim(i) = Z_sample * lambda(1:n_sample) ;
            
            if estimvar                
                Z_variance(i) = Sill * ( Gi' * lambda(1:n_sample) - lambda($) ) ;
            end
        
        end
    
        sci_PutRasterRow(handle,r,Z_estim);
    
        if  estimvar
            sci_PutRasterRow(handlevar,r,Z_variance);
        end
        
    end       
       
endfunction
