// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function [C,M,dL] = RasterizeSegment (Segment,Header)

    // Rasterization "épaisse" d'un segment sur une grille définie par Header
    // Segment = [Xdeb , Ydeb ; Xfin , Yfin ]
   
    if Header($)==0
        offset = 0;
    else
        offset = 0.5;
    end
   
    Xdeb = (Segment(1)-Header(4))/Header(3)-offset;
    Xfin = (Segment(2)-Header(4))/Header(3)-offset;
    
    Ydeb = (Segment(3)-Header(5))/Header(3)-offset;
    Yfin = (Segment(4)-Header(5))/Header(3)-offset;
      
    u = diff(Segment,1,1)' ;   
   
    // Identification des intersections entre le segment et la grille
    
    grid_X = ceil(min(Xdeb,Xfin)):floor(max(Xdeb,Xfin)) ;
    
    if (Xfin~=Xdeb & ~isempty(grid_X)) then
        t_X = (grid_X - Xdeb) / (Xfin - Xdeb) ;
    else
        t_X = [];
    end

    grid_Y = ceil(min(Ydeb,Yfin)):floor(max(Ydeb,Yfin)) ;
   
    if (Yfin~=Ydeb & ~isempty(grid_Y)) then
        t_Y = (grid_Y - Ydeb) / (Yfin - Ydeb) ;
    else
        t_Y = [];
    end
    
    t = gsort([t_X,t_Y],'g','i');
    
    if ~isempty(t) then
        M = u .*. [ 0.5*([0,t]+[t,1]) ] + [Segment(1);Segment(3)].*.ones(1,length(t)+1) ;     // Milieux des sous-segments
		M = M';
		dL = norm(u) * diff([0,t,1]');		
    else
//        M = u .*. [ 0 , 1 ] + [Segment(1);Segment(3)].*.ones(1,2) ;
	      M = mean(Segment,'r');
		  dL = norm(u);
    end
    
    C = XY2rec ( Header,M);
        
endfunction
