// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function Z_estim = RegressionKriging ( XY_sample , Z_sample , AUX_sample , XY_estim , AUX_estim , BETA , VARIO )
    
    // VARIO : list(type,Sill,Range,Nugget)
    
    n_sample = length(Z_sample) ;   
    n_aux = size(AUX_sample,'c') ;
    
    if length(BETA)<n_aux then
        error();
    end
    
    //////////////////////////////////////////////////////////////////////////// 
    /////////////// Calcul des résidus aux points échantillonnés ///////////////
    //////////////////////////////////////////////////////////////////////////// 
    
    if length(BETA)~=(n_aux+1) then    // Pas d'intercept spécifié dans le modèle déterministe
    
        R_sample = AUX_sample * BETA - Z_sample ; 
        BETA($+1) = -1 * mean ( R_sample ) ;
        R_sample = R_sample + BETA($) ;    // Les résidus aux points échantillonnés sont d'espérance nulle 
        
    else    // On garde l'intercept spécifié dans le modèle déterministe
        
        R_sample = AUX_sample * BETA(1:n_aux) + BETA(n_aux+1) - Z_sample ; 
            
    end
    
    R_sample = R_sample' ;
    
    //////////////////////////////////////////////////////////////////////////// 
    //////// Matrice de covariance des résidus aux points échantillonnés ///////
    ////////////////////////////////////////////////////////////////////////////     
    
    X_sample = XY_sample(:,1);
    Y_sample = XY_sample(:,2);
          
    DX = X_sample'.*.ones(X_sample)-X_sample.*.ones(X_sample');
    DY = Y_sample'.*.ones(Y_sample)-Y_sample.*.ones(Y_sample');
    D = sqrt(DX.^2+DY.^2);
      
    type_vario = convstr ( part(VARIO(1),1:1) , 'l' );
    Sill = VARIO(2);
    Range = VARIO(3);
    Nugget = VARIO(4);
    
    gr0 = Nugget/Sill ;
      
    select type_vario
    case 'e' then    // variogramme exponentiel
        gr = gr0 + (1-gr0) * (1-exp(-D/Range));
    case 's' then    // variogramme sphérique
        gr = ones(D);
        gr(D<Range) = gr0 + (1-gr0) * [ 1.5*D(D<Range)/Range-0.5*(D(D<Range)/Range).^3 ] ;
    case 'e' then    // variogramme gaussien
        gr = gr0 + (1-gr0) * (1-exp(-[D/Range].^2));
    end
      
    gr = gr - diag(diag(gr));
       
    R_estim = zeros ( size(XY_estim,'r') , 1 ) ;
    
    //////////////////////////////////////////////////////////////////////////// 
    ////////////////////// Estimation aux points XY_estim //////////////////////
    ////////////////////////////////////////////////////////////////////////////    

    for i=1:size(XY_estim,'r') 
        
        // Calcul de la matrice de covariance des résidus à la localité XY_estim(i,:)
        
        D = sqrt ( ( X_sample - XY_estim(i,1) ).^2 + ( Y_sample - XY_estim(i,2) ).^2 );

        select type_vario
        case 'e' then    // variogramme exponentiel
            gri = gr0 + (1-gr0) * (1-exp(-D/Range));
        case 's' then    // variogramme sphérique
            gri = ones(D);
            gri(D<Range) = gr0 + (1-gr0) * [ 1.5*D(D<Range)/Range-0.5*(D(D<Range)/Range).^3 ] ;
        case 'e' then    // variogramme gaussien
            gri = gr0 + (1-gr0) * (1-exp(-[D/Range].^2)); 
        end
      
        // Calcul des poids de krigeage
      
        A=[gr,-ones(n_sample,1);ones(1,n_sample),0];
        B=[gri;1];
        
        lambda = A\B ;
        
        R_estim(i) = R_sample * lambda(1:n_sample);
        
    end       
        
    Z_estim = AUX_estim * BETA(1:n_aux) + BETA(n_aux+1) + R_estim ;
       
endfunction
