// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function [Inum,Iden] = SegmentGradientIntegral ( XYdeb, XYfin, F, a )
    
    Vec = XYfin-XYdeb ;
    L = sqrt(sum(Vec.^2,'c')) ;
    L(L==0) = %nan ;
    
    F00 = F(:,1);
    Fa0 = F(:,2);
    F0a = F(:,3);
    Faa = F(:,4);
    
    B = (F00 + Faa - F0a - Fa0)/(a^2) ;
    Cx = (Fa0-F00)/a ;
    Cy = (F0a-F00)/a ;
    
    iB = (B~=0);
    
    // Int ||grad F|| 
    
    Iden = zeros(L);

    B(~iB) = %nan ;
    costheta = Vec(:,1)./L ;
    sintheta = Vec(:,2)./L ;
    
    Nx = B.*XYdeb(:,2) + Cx ;
    Ny = B.*XYdeb(:,1) + Cy ;
    
    ALPHA = 2*(Nx.*sintheta+Ny.*costheta)./B ;
    BETA = ( Nx.^2 + Ny.^2 ) .* (B.^-2) ;
    
    GAMMA = BETA - 0.25*ALPHA.^2 ;
    
    Binf = ALPHA/2 ;
    Bsup = L+ALPHA/2 ;
    
    Iden = 0.5*abs(B).* [ + Bsup.*sqrt(Bsup.^2 + GAMMA) + GAMMA.*log(abs(Bsup+sqrt(Bsup.^2 + GAMMA))) ..
                          - Binf.*sqrt(Binf.^2 + GAMMA) - GAMMA.*log(abs(Binf+sqrt(Binf.^2 + GAMMA))) ] ;
                  
    Iden(~iB) = sqrt(Cx(~iB).^2 + Cy(~iB).^2).*L(~iB) ;

    // Int |(grad F) . u|
    
    Inum = zeros(L) ;
    
    Fdeb = [ F00.*(a-XYdeb(:,1)).*(a-XYdeb(:,2)) + ..
             Fa0.*XYdeb(:,1).*(a-XYdeb(:,2)) + ..
             F0a.*(a-XYdeb(:,1)).*XYdeb(:,2) + ..
             Faa.*XYdeb(:,1).*XYdeb(:,2) ] / (a*a) ;

    Ffin = [ F00.*(a-XYfin(:,1)).*(a-XYfin(:,2)) + ..
             Fa0.*XYfin(:,1).*(a-XYfin(:,2)) + ..
             F0a.*(a-XYfin(:,1)).*XYfin(:,2) + ..
             Faa.*XYfin(:,1).*XYfin(:,2) ] / (a*a) ;
                
    // On voit si le produit scalaire change de signe sur le segment
    
    Bsin2theta = 2*B.*sintheta.*costheta ;
    Bsin2theta(Bsin2theta==0) = %nan ;
    
    tsgn = -(Nx.*costheta+Ny.*sintheta)./Bsin2theta ;
    
    chsgn = (tsgn>0 & tsgn<L);
    
    Inum(~chsgn) = abs(Fdeb(~chsgn)-Ffin(~chsgn)) ;
    Inum(~chsgn) = abs(Fdeb(~chsgn)-Ffin(~chsgn)) ;

    if or(chsgn) then
        
        XYsgn = zeros(XYdeb(chsgn,:));
        
        XYsgn(:,1) = XYdeb(chsgn,1) + tsgn(chsgn).*Vec(find(chsgn),1)./L(chsgn) ; 
        XYsgn(:,2) = XYdeb(chsgn,2) + tsgn(chsgn).*Vec(find(chsgn),2)./L(chsgn) ; 

        Fsgn = [ F00(chsgn).*(a-XYsgn(:,1)).*(a-XYsgn(:,2)) + ..
                 Fa0(chsgn).*XYsgn(:,1).*(a-XYsgn(:,2)) + ..
                 F0a(chsgn).*(a-XYsgn(:,1)).*XYsgn(:,2) + ..
                 Faa(chsgn).*XYsgn(:,1).*XYsgn(:,2) ] / (a*a) ;
        
        Inum(chsgn) = abs(Fdeb(chsgn)-Fsgn) + abs(Ffin(chsgn)-Fsgn) ;

    end

    Inum(isnan(Inum))=0;
    Iden(isnan(Iden))=0;

endfunction
