// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function [ VARIO_ResiTemplateP ] = TemplateP_variography ( PRECIP , XYZ_precip , ScaleHeight )
    
    // PRECIP            :     Matrix [ndays,nprecip]    :    Daily rainfall time series at gauges
    // XYZ_precip        :     Matrix [nprecip,3]        :    X,Y,Z coordinates of rain gauges
    // ScaleHeight       :     Integer                   
    // MeanVariancePrecip :    Vector [4]                :    Parameters of the mean-variance relationship of the local scaling factor
    
    PRECIP(PRECIP<0) = %nan ;
    
    if or(isnan(PRECIP)) then
        error("Les series pluviometriques ne doivent pas comporter de lacunes");
    end
    
    [ndays,nprecip] = size(PRECIP);
    
    if or( size(XYZ_precip) ~= [nprecip,3] ) then
        error("Les arguments 1 et 3 ont des dimensions incompatibles");
    end

    //////////////////////////////////////////////////////////////////////////// 
    ///////////// Compute sample-sample symmetric distance matrices ////////////
    ////////////////////////////////////////////////////////////////////////////
    
    DX = XYZ_precip(:,1)'.*.ones(nprecip,1)-XYZ_precip(:,1).*.ones(1,nprecip);
    DY = XYZ_precip(:,2)'.*.ones(nprecip,1)-XYZ_precip(:,2).*.ones(1,nprecip);
    D_pp = sqrt(DX.^2+DY.^2);
                        
    //////////////////////////////////////////////////////////////////////////// 
    ///////// Build precipitation and temperature template at stations /////////
    ////////////////////////////////////////////////////////////////////////////            
       
    TemplateP_sample = mean(PRECIP,'r') ;

    ResiTemplateP_sample = log(TemplateP_sample) - XYZ_precip(:,3)'/ScaleHeight;
    P0 = exp(mean(ResiTemplateP_sample));
    ResiTemplateP_sample = ResiTemplateP_sample - mean(ResiTemplateP_sample);  
    
    //////////////////////////////////////////////////////////////////////////// 
    //////// Variographie des résidus à la dérive (intermède musical 1) ////////
    ////////////////////////////////////////////////////////////////////////////            
   
    [ h_window , g , D ] = VarioEmpirique ( XYZ_precip(:,1:2) , ResiTemplateP_sample , nprecip , %T ) ;
    h = mean ( h_window , 'c' ) ;        
    Sill = variance(ResiTemplateP_sample) ;                       
    Nugget = 0 ;
    X0 = [ 0 ] ;
    [fopt,Xopt,gropt] = leastsq ( 0 , list ( VarioDiff_wo_nugget , Nugget , Sill , 'spherical' , h , g ) , X0 , 'ar' , 100 , 50 );        
    Range = max(h) * 0.5 * ( 1 + tanh ( Xopt(1) ) );
        
   // Structural function for precipitation and temperature templates
    
    VARIO_ResiTemplateP = list ( "spherical" , Sill , Range , Nugget );
    
endfunction
