// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function [ VARIO_ResiTemplateT ] = TemplateT_variography ( TEMPER , XYZ_temper , LapseRate )
    
    
    if or(isnan(TEMPER)) then
        error("Les series pluviometriques ne doivent pas comporter de lacunes");
    end
    
    [ndays,ntemper] = size(TEMPER);
    
    if or( size(XYZ_temper) ~= [ntemper,3] ) then
        error("Les arguments 1 et 2 ont des dimensions incompatibles");
    end

    //////////////////////////////////////////////////////////////////////////// 
    ///////////// Compute sample-sample symmetric distance matrices ////////////
    ////////////////////////////////////////////////////////////////////////////

    DX = XYZ_temper(:,1)'.*.ones(ntemper,1)-XYZ_temper(:,1).*.ones(1,ntemper);
    DY = XYZ_temper(:,2)'.*.ones(ntemper,1)-XYZ_temper(:,2).*.ones(1,ntemper);
    D_tt = sqrt(DX.^2+DY.^2);
                        
    //////////////////////////////////////////////////////////////////////////// 
    ///////// Build precipitation and temperature template at stations /////////
    ////////////////////////////////////////////////////////////////////////////            
       
    TemplateT_sample = mean(TEMPER,'r') ;

    ResiTemplateT_sample = TemplateT_sample - LapseRate * XYZ_temper(:,3)';
    T0 = mean(ResiTemplateT_sample);
    ResiTemplateT_sample = ResiTemplateT_sample - T0 ;  
    
    //////////////////////////////////////////////////////////////////////////// 
    //////// Variographie des résidus à la dérive (intermède musical 1) ////////
    ////////////////////////////////////////////////////////////////////////////            
   
    [ h_window , g , D ] = VarioEmpirique ( XYZ_temper(:,1:2) , ResiTemplateT_sample , ntemper , %T ) ;
    h = mean ( h_window , 'c' ) ;        
    Sill = variance(ResiTemplateT_sample) ;                       
    Nugget = 0 ;
    X0 = [ 0 ] ;
    [fopt,Xopt,gropt] = leastsq ( 0 , list ( VarioDiff_wo_nugget , Nugget , Sill , 'spherical' , h , g ) , X0 , 'ar' , 100 , 50 );        
    Range = max(h) * 0.5 * ( 1 + tanh ( Xopt(1) ) );
        
   // Structural function for precipitation and temperature templates
    
    VARIO_ResiTemplateT = list ( "spherical" , Sill , Range , Nugget );
    
endfunction
