// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function [h,g] = VarioDeformeEbauche ( XY , PJ , Ebauche )
    
    X = XY(:,1);
    Y = XY(:,2);
    n = length(X);
    T = triu(ones(n,n));
    P1 = [1:n].*.ones(n,1);
    P2 = ones(1,n).*.[1:n]';
    
    paires = find(T==0);
          
    DX = X'.*.ones(X)-X.*.ones(X');
    DY = Y'.*.ones(Y)-Y.*.ones(Y');   
    D = sqrt(DX.^2+DY.^2);

    DX = DX(paires);
    DY = DY(paires);
    h = D(paires);
    P1 = P1(paires);
    P2 = P2(paires);

    Rj = zeros(PJ) ;
    XY0 = XY;
    XY0(:,1) = [XY(:,1)-mean(XY(:,1))]/stdev(XY(:,1));
    XY0(:,2) = [XY(:,2)-mean(XY(:,2))]/stdev(XY(:,2));

    for j=1:size(PJ,'r')
                  
        // Pour chaque journée j, on calcule la transfo affine optimale telle que :
        // ChampJ(xi,yi,j) = max { a(j) * Ebauche(xi,yi) + b(j) ; 0 } + Rj(xi,yi)
        // pour les (xi,yi) du panier de postes
    
        A = [ Ebauche(:) , ones(Ebauche(:))  , XY0 ] ;
        X = PJ(j,:)' ;
        DeformeAffine = A\X ;
        [fopt,DeformeAffine,gropt] = leastsq ( 0 , list ( LineaireSeuil , X , A ) , DeformeAffine , 'ar' , 100 , 100 );
        R = max(A * DeformeAffine,0) - X ;
//        plot(X,[Ebauche(:),max(A * DeformeAffine,0)],'.');
//        plot([0 max(X)],[0 max(X)],'--k');

        if  sum(X>0)>1 then
       
            m = mean(log(X(X>0)));
            s = stdev(log(X(X>0)));
            F0 = sum(X==0)/length(X);
            
            F = ( [1:length(X)] - 3/8 ) / (length(X)+1-2*3/8) ;
            Xs = gsort(X,'g','i');
            Xg = 0.1:0.1:Xs($);
            plot([0,Xg],[F0,F0+0.5*(1-F0)*erfc(-(log(Xg)-m)/(s*sqrt(2)))]) ; 
            plot(Xs,F,'.')
        
        end        

        pause
        
        if stdev(R)>0
            Rj(j,:) = R' / stdev(R);
        else
            Rj(j,:) = 0;
        end
                    
    end


    for i=1:length(h)
              
        // On considère chaque champ de résidus journaliers normalisés Rj(x,y)
        // comme une réalisation de la même loi R(x,y)
        // (c'est l'hypothèse que l'on fait en appliquant le même vario)
        // => on peut évaluer une variance pour chaque paire de points       
        
        E = Rj(:,P1(i)) - Rj(:,P2(i)) ;
        g(i) = variance(E);
                                
    end

endfunction
