// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function W = WidthFunction(tree,rast,a0,V)
    
    // ARGUMENTS :
    // tree : identifiant d'un objet de type CellTree
    // rast : identifiant de la grille d'altitude ou de distance hydraulique
    // a0 : seuil d'extraction pour la width function
    // V : vecteur des altitutes ou de distances hydrauliques auxquelles on veut une évaluation de la width function
    
    // SORTIES :
    // W : width function évaluée aux points de V
    
    W = zeros(V);
    
    [ STREAMS , downStream , downCell ] = MakeStreamLines ( tree , a0 , %F );
    
    for i=1:length(STREAMS)
        
        Bi = GetRasterValue(rast,[ STREAMS(i)($) , downCell(i) ]);
        Bi = gsort(Bi,'g','i');
        
        // Courbes de niveaux intersectées par le drain i :
        
        ix = (V>Bi(1)) & (V<Bi(2)) ;
        W(ix) = W(ix)+1;
        
    end
    
endfunction
