// 
// Copyright (C) 2013 UMPC
// 
// This file is part of CLaRiNet.
//  
// CLaRiNet is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// CLaRiNet is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with CLaRiNet.  If not, see <http://www.gnu.org/licenses/>.
function V = bilinear ( HandleRaster , XY )
    
    Header = GetFloatRasterHeader(HandleRaster);
    ncols = Header(1);
    nrows = Header(2);
    
    a = Header(3)
    GridOrigin = [Header(4),Header(5)];
    
    if Header($)==0    // Si (x,y)llcorner
        GridOrigin = GridOrigin + a/2 ;
    end
        
    OR = GridOrigin.*.ones(size(XY,'r'),1) ;
    offXY = a * floor ( (1/a)*(XY - OR) ) ;
    xy = XY-OR-offXY ;
    
    HeaderT = Header ;
    HeaderT(4:5) = HeaderT(4:5)+a/2 ;
    
    RC00 = XY2RC(HeaderT,XY);
    row00 = RC00(:,1);
    col00 = RC00(:,2);
    
    rowa0 = row00;
    row0a = row00-1;
    rowaa = row00-1;

    cola0 = col00+1;
    col0a = col00;
    colaa = col00+1;
    
    COL = [col00,cola0,col0a,colaa];
    ROW = [row00,rowa0,row0a,rowaa];
           
    ROW(ROW<0|ROW>nrows) = %nan;
    COL(COL<0|COL>ncols) = %nan;  
    
    REC = ROW*ncols + COL ;
    isin = ~or(isnan(REC),'c') ;
    
    V = zeros(size(XY,'r'),1);
    
    V(~isin) = %nan ;
    isin = find(isin);
            
    F = GetRasterValue ( HandleRaster , REC(isin,:) ) ;
    F(F<=Header($-1)) = %nan ;

    x = xy(isin,1);
    y = xy(isin,2);
    
    V(isin)= 1/(a^2) * [ F(:,1).*(a-x).*(a-y) + F(:,2).*x.*(a-y) + F(:,3).*(a-x).*y + F(:,4).*x.*y ];
    
endfunction
